Package: magi
Type: Package
Title: MAnifold-Constrained Gaussian Process Inference
Version: 1.2.4
Date: 2024-06-21
Authors@R: 
  c(person(given = "Shihao",
           family = "Yang",
           role = c("aut", "cre"),
           email = "shihao.yang@isye.gatech.edu",
           comment = c(ORCID = "0000-0003-3910-4969")),
    person(given = "Samuel W.K.",
           family = "Wong",
           role = "aut",
           comment = c(ORCID = "0000-0002-7325-7267")),
    person(given = "S.C.",
           family = "Kou",
           role = c("ctb", "cph"),
           comment = c(role = "Contributor of MAGI method development"))
           )
Encoding: UTF-8
Description: Provides fast and accurate inference for the parameter estimation problem in Ordinary Differential
    Equations, including the case when there are unobserved system components. Implements the MAGI method
    (MAnifold-constrained Gaussian process Inference) of Yang, Wong, and Kou (2021) <doi:10.1073/pnas.2020397118>.
    A user guide is provided by the accompanying software paper Wong, Yang, and Kou (2024) <doi:10.18637/jss.v109.i04>.
URL: https://doi.org/10.18637/jss.v109.i04
License: MIT + file LICENSE
VignetteBuilder: knitr
Imports: Rcpp (>= 1.0.6), gridExtra, gridBase, grid, methods, deSolve
LinkingTo: Rcpp, RcppArmadillo, BH, roptim
RoxygenNote: 7.3.1
Suggests: testthat, mvtnorm, covr, knitr, MASS, rmarkdown, markdown
Depends: R (>= 3.6.0)
NeedsCompilation: yes
Packaged: 2024-06-21 21:13:19 UTC; s246wong
Author: Shihao Yang [aut, cre] (<https://orcid.org/0000-0003-3910-4969>),
  Samuel W.K. Wong [aut] (<https://orcid.org/0000-0002-7325-7267>),
  S.C. Kou [ctb, cph] (Contributor of MAGI method development)
Maintainer: Shihao Yang <shihao.yang@isye.gatech.edu>
Repository: CRAN
Date/Publication: 2024-06-22 17:30:02 UTC
Built: R 4.4.1; x86_64-apple-darwin12.6.0; 2024-06-23 18:26:30 UTC; unix
