# Generated by roxygen2: do not edit by hand

S3method(plot,magioutput)
S3method(print,magioutput)
S3method(summary,magioutput)
export(MagiPosterior)
export(MagiSolver)
export(calCov)
export(fnmodelDtheta)
export(fnmodelDx)
export(fnmodelODE)
export(gpcov)
export(gpmean)
export(gpsmoothing)
export(gpsmoothllik)
export(hes1logmodelDtheta)
export(hes1logmodelDx)
export(hes1logmodelODE)
export(hes1modelDtheta)
export(hes1modelDx)
export(hes1modelODE)
export(is.magioutput)
export(magioutput)
export(ptransmodelDtheta)
export(ptransmodelDx)
export(ptransmodelODE)
export(setDiscretization)
export(testDynamicalModel)
import(Rcpp)
importFrom(grDevices,dev.off)
importFrom(grDevices,pdf)
importFrom(graphics,abline)
importFrom(graphics,hist)
importFrom(graphics,layout)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,matplot)
importFrom(graphics,mtext)
importFrom(graphics,par)
importFrom(graphics,plot.function)
importFrom(graphics,plot.new)
importFrom(graphics,points)
importFrom(graphics,polygon)
importFrom(graphics,title)
importFrom(grid,pushViewport)
importFrom(gridBase,baseViewports)
importFrom(gridExtra,grid.table)
importFrom(stats,approx)
importFrom(stats,density)
importFrom(stats,dist)
importFrom(stats,dlnorm)
importFrom(stats,dnorm)
importFrom(stats,fft)
importFrom(stats,median)
importFrom(stats,nobs)
importFrom(stats,plot.ts)
importFrom(stats,quantile)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,weighted.mean)
importFrom(utils,head)
importFrom(utils,tail)
useDynLib(magi)
